#include "Navigator.h"

#include <algorithm>
#include <Eigen/src/Geometry/Transform.h>

#include <VirtualRobot/Robot.h>

#include "ArmarXCore/core/exceptions/local/ExpressionException.h"

#include "Navigation/libraries/core/types.h"

namespace armarx::nav::server
{

    // Stack result
    core::TrajectoryPtr StackResult::trajectory() const
    {
        if (localTrajectory)
        {
            return localTrajectory;
        }

        ARMARX_CHECK_NOT_NULL(globalTrajectory);
        return globalTrajectory;
    }

    core::Twist StackResult::velocity() const
    {
        if (safeVelocity.has_value())
        {
            return safeVelocity.value();
        }

        ARMARX_CHECK(controlVelocity.has_value());
        return *controlVelocity;
    }

    // Navigator

    Navigator::Navigator(const server::NavigationStack& navigationStack,
                         const core::Scene& scene,
                         ExecutorInterface& executor) :
        stack(navigationStack), scene(scene), executor{executor}
    {
        setTag("Navigator");
    }

    Navigator::~Navigator()
    {
        stop();
    }

    void Navigator::moveTo(const std::vector<core::Pose>& waypoints, const core::NavigationFrames& navigationFrame)
    {
        std::vector<core::Pose> globalWaypoints;
        switch (navigationFrame)
        {
            case core::NavigationFrames::Absolute:
                globalWaypoints = waypoints;
                break;
            case core::NavigationFrames::Relative:
                globalWaypoints.reserve(waypoints.size());
                std::transform(std::begin(waypoints), std::end(waypoints), std::back_inserter(globalWaypoints), [&](const core::Pose & p)
                {
                    return core::Pose(scene.robot->getGlobalPose()) * p;
                });
                break;
        }

        moveToAbsolute(globalWaypoints);
    }

    void Navigator::moveToAbsolute(const std::vector<core::Pose>& waypoints)
    {
        ARMARX_INFO << "Planning global trajectory";
        ARMARX_CHECK_NOT_NULL(stack.globalPlanner);

        StackResult res;
        res.globalTrajectory = stack.globalPlanner->plan(core::Pose(waypoints.at(0)));

        if (stack.localPlanner)
        {
            ARMARX_INFO << "Planning local trajectory";

            // TODO: start the local planner thread
            res.localTrajectory = stack.localPlanner->plan(res.globalTrajectory);
        }

        ARMARX_CHECK_NOT_NULL(stack.trajectoryControl);
        while (true)
        {
            ARMARX_INFO << "Trajectory control";

            res.controlVelocity = stack.trajectoryControl->control(res.trajectory());

            if (stack.safetyControl)
            {
                ARMARX_INFO << "Safety control";
                res.safeVelocity = stack.safetyControl->control(res.controlVelocity.value());
            }

            executor.move(res.velocity());
        }
    }

    void Navigator::moveTowards(const core::Direction& direction, const core::NavigationFrames& navigationFrame)
    {

    }

    void Navigator::moveTowardsAbsolute(const core::Direction& direction)
    {

    }

    void Navigator::stop()
    {

    }

} // namespace armarx::nav::server
